import json
import os
import random
import time
from abc import ABC
from abc import abstractmethod
from io import TextIOWrapper
from typing import Dict
from typing import Optional
from typing import Type
from typing import Union

from ..logs.base import Logger
from ..simulator import Simulator


class Runner(ABC):
    def __init__(
        self,
        settings: Union[Dict, TextIOWrapper, os.PathLike, str],
        prng: Optional[random.Random] = None,
        logger: Optional[Logger] = None,
        simulator_class: Type[Simulator] = Simulator,
    ):
        self.settings: Dict
        if isinstance(settings, Dict):
            self.settings = settings
        elif isinstance(settings, TextIOWrapper):
            self.settings = json.load(fp=settings)
        else:
            self.settings = json.load(fp=open(settings, mode="r"))
        self._prng: random.Random = prng if prng is not None else random.Random()
        self.logger = logger
        self.simulator: Simulator = simulator_class(
            prng=random.Random(self._prng.randint(0, 2**31))
        )

    def main(self) -> None:
        setup_start_time_ns = time.time_ns()
        self._setup()
        start_time_ns = time.time_ns()
        self._run()
        end_time_ns = time.time_ns()
        print(
            "# INITIALIZATION TIME " + str((start_time_ns - setup_start_time_ns) / 1e9)
        )
        print("# EXECUTION TIME " + str((end_time_ns - start_time_ns) / 1e9))

    @abstractmethod
    def _setup(self) -> None:
        pass

    @abstractmethod
    def _run(self) -> None:
        pass



import random
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple

import numpy as np
from scipy.linalg import cholesky


class Fundamentals:
    def __init__(self, prng: random.Random) -> None:
        self._prng = prng
        self._np_prng: np.random.Generator = np.random.default_rng(
            self._prng.randint(0, 2**31)
        )
        self.correlation: Dict[Tuple[int, int], float] = {}
        self.drifts: Dict[int, float] = {}
        self.volatilities: Dict[int, float] = {}
        self.prices: Dict[int, List[float]] = {}
        self.market_ids: List[int] = []
        self.initials: Dict[int, float] = {}
        self.start_at: Dict[int, int] = {}
        self._generated_until: int = 0
        self._generate_chunk_size = 100

    def add_market(
        self,
        market_id: int,
        initial: float,
        drift: float,
        volatility: float,
        start_at: int = 0,
    ) -> None:
        if market_id in self.market_ids:
            raise ValueError(f"market {market_id} is already registered")
        if volatility < 0.0:
            raise ValueError("volatility must be non-negative")
        if initial <= 0.0:
            raise ValueError("initial value must be positive")
        self.market_ids.append(market_id)
        self.drifts[market_id] = drift
        self.volatilities[market_id] = volatility
        self.initials[market_id] = initial
        self.start_at[market_id] = start_at
        self.prices[market_id] = [initial for _ in range(start_at + 1)]
        self._generated_until = min(start_at, self._generated_until)

    def remove_market(self, market_id: int) -> None:
        self.market_ids.remove(market_id)
        self.drifts.pop(market_id)
        self.volatilities.pop(market_id)
        self.initials.pop(market_id)
        self.start_at.pop(market_id)
        self.prices.pop(market_id)

    def change_volatility(
        self, market_id: int, volatility: float, time: int = 0
    ) -> None:
        if volatility < 0.0:
            raise ValueError("volatility must be non-negative")
        self.volatilities[market_id] = volatility
        self._generated_until = time

    def change_drift(self, market_id: int, drift: float, time: int = 0) -> None:
        self.drifts[market_id] = drift
        self._generated_until = time

    def set_correlation(
        self, market_id1: int, market_id2: int, corr: float, time: int = 0
    ) -> None:
        if not (-1.0 < corr < 1.0):
            raise ValueError("corr must be between 0.0 and 1.0")
        if market_id1 == market_id2:
            raise ValueError("market_id1 and market_id2 must be different")
        if (market_id2, market_id1) in self.correlation:
            self.correlation[(market_id2, market_id1)] = corr
        else:
            self.correlation[(market_id1, market_id2)] = corr
        self._generated_until = time

    def remove_correlation(
        self, market_id1: int, market_id2: int, time: int = 0
    ) -> None:
        if market_id1 == market_id2:
            raise ValueError("market_id1 and market_id2 must be different")
        if (market_id2, market_id1) in self.correlation:
            self.correlation.pop((market_id2, market_id1))
        else:
            self.correlation.pop((market_id1, market_id2))
        self._generated_until = time

    def _generate_log_return(
        self, generate_target_ids: List[int], length: int
    ) -> np.ndarray:
        generate_target_ids_cholesky = list(
            filter(lambda x: self.volatilities[x] != 0.0, generate_target_ids)
        )
        generate_target_ids_other = list(
            filter(lambda x: self.volatilities[x] == 0.0, generate_target_ids)
        )
        corr_matrix = np.eye(len(generate_target_ids_cholesky))
        for (id1, id2), corr in self.correlation.items():
            if id1 not in generate_target_ids_cholesky:
                continue
            if id2 not in generate_target_ids_cholesky:
                continue
            if id1 == id2:
                raise AssertionError
            corr_matrix[
                generate_target_ids_cholesky.index(id1),
                generate_target_ids_cholesky.index(id2),
            ] = corr
            corr_matrix[
                generate_target_ids_cholesky.index(id2),
                generate_target_ids_cholesky.index(id1),
            ] = corr
        vol = np.asarray([self.volatilities[x] for x in generate_target_ids_cholesky])
        cov_matrix = vol * corr_matrix * vol.reshape(-1, 1)
        cholesky_matrix = cholesky(cov_matrix, lower=True)

        dw_cholesky = self._np_prng.standard_normal(
            size=(len(generate_target_ids_cholesky), length)
        )
        drifts_cholesky = np.asarray(
            [self.drifts[x] for x in generate_target_ids_cholesky]
        )
        result_cholesky = np.dot(
            cholesky_matrix, dw_cholesky
        ) + drifts_cholesky.T.reshape(-1, 1)

        drifts_others = np.asarray(
            [[self.drifts[x] for _ in range(length)] for x in generate_target_ids_other]
        )

        return np.stack(
            [
                result_cholesky[generate_target_ids_cholesky.index(x)]
                if x in generate_target_ids_cholesky
                else drifts_others[generate_target_ids_other.index(x)]
                for x in generate_target_ids
            ]
        )

    def _generate_next(self) -> None:
        setting_change_points: List[int] = [
            x for x in self.start_at.values() if x > self._generated_until
        ]
        if len(setting_change_points) == 0:
            length = self._generate_chunk_size
        else:
            length = min(setting_change_points) - self._generated_until
        next_until = self._generated_until + length
        target_market_ids: List[int] = [
            key for key, value in self.start_at.items() if value < next_until
        ]
        log_return = self._generate_log_return(
            generate_target_ids=target_market_ids, length=length
        )
        current_prices = np.asarray(
            [self.prices[x][self._generated_until] for x in target_market_ids]
        )
        prices = current_prices.T.reshape(-1, 1) * np.exp(
            np.cumsum(log_return, axis=-1)
        )
        for market_id, price_seq in zip(target_market_ids, prices):
            self.prices[market_id] = (
                self.prices[market_id][: self._generated_until + 1] + price_seq.tolist()
            )
        self._generated_until += length

    def get_fundamental_price(self, market_id: int, time: int) -> float:
        while time >= self._generated_until:
            self._generate_next()
        return self.prices[market_id][time]

    def get_fundamental_prices(
        self, market_id: int, times: Iterable[int]
    ) -> List[float]:
        while max([x for x in times]) >= self._generated_until:
            self._generate_next()
        return [self.prices[market_id][x] for x in times]



import math
import random
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from ..logs.base import Logger
from ..market import Market
from ..order import LIMIT_ORDER
from ..order import Cancel
from ..order import Order
from ..utils.json_random import JsonRandom
from .base import Agent

MARGIN_FIXED = 0
MARGIN_NORMAL = 1


class FCNAgent(Agent):
    fundamental_weight: float
    chart_weight: float
    is_chart_following: bool = True
    margin_type: int
    mean_reversion_time: int
    noise_scale: float
    noise_weight: float
    order_margin: float
    time_window_size: int

    def __init__(
        self,
        agent_id: int,
        prng: random.Random,
        simulator: "Simulator",  # type: ignore
        name: str,
        logger: Optional[Logger] = None,
    ):
        super().__init__(agent_id, prng, simulator, name, logger)

    def is_finite(self, x: float) -> bool:
        return not math.isnan(x) and not math.isinf(x)

    def setup(
        self,
        settings: Dict[str, Any],
        accessible_markets_ids: List[int],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().setup(settings=settings, accessible_markets_ids=accessible_markets_ids)
        json_random: JsonRandom = JsonRandom(prng=self.prng)
        self.fundamental_weight = json_random.random(
            json_value=settings["fundamentalWeight"]
        )
        self.chart_weight = json_random.random(json_value=settings["chartWeight"])
        self.noise_weight = json_random.random(json_value=settings["noiseWeight"])
        self.noise_scale = json_random.random(json_value=settings["noiseScale"])
        self.time_window_size = int(
            json_random.random(json_value=settings["timeWindowSize"])
        )
        self.order_margin = json_random.random(json_value=settings["orderMargin"])
        if settings.get("marginType") in [None, "fixed"]:
            self.margin_type = MARGIN_FIXED
        else:
            self.margin_type = MARGIN_NORMAL
        if "meanReversionTime" in settings:
            self.mean_reversion_time = int(
                json_random.random(json_value=settings["meanReversionTime"])
            )
        else:
            self.mean_reversion_time = self.time_window_size

    def submit_orders(self, markets: List[Market]) -> List[Union[Order, Cancel]]:
        orders: List[Union[Order, Cancel]] = sum(
            [self.submit_orders_by_market(market=market) for market in markets], []
        )
        return orders

    def submit_orders_by_market(self, market: Market) -> List[Union[Order, Cancel]]:
        orders: List[Union[Order, Cancel]] = []
        if not self.is_market_accessible(market_id=market.market_id):
            return orders

        time: int = market.get_time()
        time_window_size: int = min(time, self.time_window_size)
        assert time_window_size >= 0
        assert self.fundamental_weight >= 0.0
        assert self.chart_weight >= 0.0
        assert self.noise_weight >= 0.0

        fundamental_scale: float = 1.0 / max(self.mean_reversion_time, 1)
        fundamental_log_return = fundamental_scale * math.log(
            market.get_fundamental_price() / market.get_market_price()
        )
        assert self.is_finite(fundamental_log_return)

        chart_scale: float = 1.0 / max(time_window_size, 1)
        chart_mean_log_return = chart_scale * math.log(
            market.get_market_price() / market.get_market_price(time - time_window_size)
        )
        assert self.is_finite(chart_mean_log_return)

        noise_log_return: float = self.noise_scale * self.prng.gauss(mu=0.0, sigma=1.0)
        assert self.is_finite(noise_log_return)

        expected_log_return: float = (
            1.0 / (self.fundamental_weight + self.chart_weight + self.noise_weight)
        ) * (
            self.fundamental_weight * fundamental_log_return
            + self.chart_weight
            * chart_mean_log_return
            * (1 if self.is_chart_following else -1)
            + self.noise_weight * noise_log_return
        )
        assert self.is_finite(expected_log_return)

        expected_future_price: float = market.get_market_price() * math.exp(
            expected_log_return * self.time_window_size
        )
        assert self.is_finite(expected_future_price)

        if self.margin_type == MARGIN_FIXED:
            assert 0.0 <= self.order_margin <= 1.0

            order_volume: int = 1

            if expected_future_price > market.get_market_price():
                order_price = expected_future_price * (1 - self.order_margin)
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=market.market_id,
                        is_buy=True,
                        kind=LIMIT_ORDER,
                        volume=order_volume,
                        price=order_price,
                        ttl=self.time_window_size,
                    )
                )
            if expected_future_price < market.get_market_price():
                order_price = expected_future_price * (1 + self.order_margin)
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=market.market_id,
                        is_buy=False,
                        kind=LIMIT_ORDER,
                        volume=order_volume,
                        price=order_price,
                        ttl=self.time_window_size,
                    )
                )

        if self.margin_type == MARGIN_NORMAL:
            assert self.order_margin >= 0.0
            order_price = (
                expected_future_price
                + self.prng.gauss(mu=0.0, sigma=1.0) * self.order_margin
            )
            order_volume = 1
            assert order_price >= 0.0
            assert order_volume > 0
            if expected_future_price > market.get_market_price():
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=market.market_id,
                        is_buy=True,
                        kind=LIMIT_ORDER,
                        volume=order_volume,
                        price=order_price,
                        ttl=self.time_window_size,
                    )
                )
            if expected_future_price < market.get_market_price():
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=market.market_id,
                        is_buy=False,
                        kind=LIMIT_ORDER,
                        volume=order_volume,
                        price=order_price,
                        ttl=self.time_window_size,
                    )
                )
        return orders

    def __str__(self) -> str:
        return (
            f"Agent:{self.agent_id}[rnd:{self.prng},cw:{self.chart_weight},fw:{self.fundamental_weight},"
            + f"following:{self.is_chart_following},mtypr:{self.margin_type},mrt:{self.mean_reversion_time},"
            + f"ns:{self.noise_scale},nw:{self.noise_weight},om:{self.order_margin},tws:{self.time_window_size},"
            + f"cash:{self.cash_amount}]"
        )



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from ..index_market import IndexMarket
from ..market import Market
from ..order import LIMIT_ORDER
from ..order import Cancel
from ..order import Order
from .high_frequency_agent import HighFrequencyAgent


class ArbitrageAgent(HighFrequencyAgent):
    order_volume: int = 1
    order_threshold_price: float = 1.0
    order_time_length: int = 1

    def setup(  # type: ignore
        self,
        settings: Dict[str, Any],
        accessible_markets_ids: List[int],
        *args,
        **kwargs
    ) -> None:
        super(ArbitrageAgent, self).setup(
            settings, accessible_markets_ids, *args, **kwargs
        )
        if "orderVolume" not in settings:
            raise ValueError("orderVolume is required for ArbitrageAgent")
        self.order_volume = settings["orderVolume"]
        if "orderThresholdPrice" not in settings:
            raise ValueError("orderThresholdPrice is required for ArbitrageAgent")
        self.order_threshold_price = settings["orderThresholdPrice"]
        if "orderTimeLength" in settings:
            self.order_time_length = settings["orderTimeLength"]

    def _submit_orders(self, market: Market) -> List[Union[Order, Cancel]]:
        orders: List[Union[Order, Cancel]] = []
        if not isinstance(market, IndexMarket):
            return orders
        if not self.is_market_accessible(market_id=market.market_id):
            return orders
        index: IndexMarket = market
        spots: List[Market] = index.get_components()
        if not index.is_running or not index.is_all_markets_running():
            return orders
        market_index: float = index.get_index()
        market_price: float = index.get_market_price()

        if len(set(map(lambda x: x.outstanding_shares, spots))) > 1:
            raise AssertionError(
                "currently, the components must have the same outstanding shares"
            )

        if (
            market_price < market_index
            and market_index - market_price > self.order_threshold_price
        ):
            index_order_volume = len(spots) * self.order_volume
            orders.append(
                Order(
                    agent_id=self.agent_id,
                    market_id=index.market_id,
                    is_buy=True,
                    kind=LIMIT_ORDER,
                    volume=index_order_volume,
                    price=index.get_market_price(),
                    ttl=self.order_time_length,
                )
            )
            for m in spots:
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=m.market_id,
                        is_buy=False,
                        kind=LIMIT_ORDER,
                        volume=self.order_volume,
                        price=m.get_market_price(),
                        ttl=self.order_time_length,
                    )
                )
        if (
            market_price > market_index
            and market_price - market_index > self.order_threshold_price
        ):
            index_order_volume = len(spots) * self.order_volume
            orders.append(
                Order(
                    agent_id=self.agent_id,
                    market_id=index.market_id,
                    is_buy=False,
                    kind=LIMIT_ORDER,
                    volume=index_order_volume,
                    price=index.get_market_price(),
                    ttl=self.order_time_length,
                )
            )
            for m in spots:
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=m.market_id,
                        is_buy=True,
                        kind=LIMIT_ORDER,
                        volume=self.order_volume,
                        price=m.get_market_price(),
                        ttl=self.order_time_length,
                    )
                )
        return orders

    def submit_orders(self, markets: List[Market]) -> List[Union[Order, Cancel]]:
        orders: List[Union[Order, Cancel]] = []
        for market in markets:
            orders.extend(self._submit_orders(market=market))
        return orders



import math
import random
from typing import Any
from typing import Dict
from typing import Final
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast

from .logs.base import CancelLog
from .logs.base import ExecutionLog
from .logs.base import Log
from .logs.base import Logger
from .logs.base import OrderLog
from .order import Cancel
from .order import Order
from .order_book import OrderBook

T = TypeVar("T")


class Market:
    def __init__(
        self, market_id: int, prng: random.Random, simulator: "Simulator", name: str, logger: Optional[Logger] = None  # type: ignore
    ) -> None:
        self.market_id: int = market_id
        self._prng = prng
        self.logger: Optional[Logger] = logger
        self._is_running: bool = False
        self.tick_size: float = 1.0
        self.chunk_size = 100
        self.sell_order_book: OrderBook = OrderBook(is_buy=False)
        self.buy_order_book: OrderBook = OrderBook(is_buy=True)
        self.time: int = -1
        self._market_prices: List[Optional[float]] = []
        self._last_executed_prices: List[Optional[float]] = []
        self._mid_prices: List[Optional[float]] = []
        self._fundamental_prices: List[Optional[float]] = []
        self._executed_volumes: List[int] = []
        self._executed_total_prices: List[float] = []
        self._n_buy_orders: List[int] = []
        self._n_sell_orders: List[int] = []
        self._next_order_id: int = 0
        self._simulator: "Simulator" = simulator  # type: ignore
        self.name: str = name
        self.outstanding_shares: Optional[int] = None

    def setup(self, settings: Dict[str, Any], *args, **kwargs) -> None:  # type: ignore
        if "tickSize" not in settings:
            raise AssertionError("tickSize is required")
        self.tick_size = settings["tickSize"]
        if "outstandingShares" in settings:
            self.outstanding_shares = settings["outstandingShares"]
        if "marketPrice" in settings:
            self._market_prices = [float(settings["marketPrice"])]
        elif "fundamentalPrice" in settings:
            self._market_prices = [float(settings["fundamentalPrice"])]
        else:
            raise ValueError(f"fundamentalPrice or marketPrice is required for market")

    def _extract_sequential_data_by_time(
        self,
        times: Union[Iterable[int], None],
        parameters: List[Optional[T]],
        allow_none: bool = False,
    ) -> List[Optional[T]]:
        if times is None:
            times = range(self.time + 1)
        if sum([t > self.time for t in times]) > 0:
            raise AssertionError("Cannot refer the future parameters")
        result = [parameters[t] for t in times]
        if not allow_none and None in result:
            raise AssertionError
        return result

    def _extract_data_by_time(
        self,
        time: Union[int, None],
        parameters: List[Optional[T]],
        allow_none: bool = False,
    ) -> Optional[T]:
        if time is None:
            time = self.time
        if time > self.time:
            raise AssertionError("Cannot refer the future parameters")
        result = parameters[time]
        if not allow_none and result is None:
            raise AssertionError
        return result

    def get_time(self) -> int:
        return self.time

    def get_market_prices(
        self, times: Union[Iterable[int], None] = None
    ) -> List[float]:
        """last executed, midprice, fundamental"""
        return cast(
            List[float],
            self._extract_sequential_data_by_time(
                times, self._market_prices, allow_none=False
            ),
        )

    def get_market_price(self, time: Union[int, None] = None) -> float:
        return cast(
            float,
            self._extract_data_by_time(time, self._market_prices, allow_none=False),
        )

    def get_mid_prices(
        self, times: Union[Iterable[int], None] = None
    ) -> List[Optional[float]]:
        return self._extract_sequential_data_by_time(
            times, self._mid_prices, allow_none=True
        )

    def get_mid_price(self, time: Union[int, None] = None) -> Optional[float]:
        return self._extract_data_by_time(time, self._mid_prices, allow_none=False)

    def get_last_executed_prices(
        self, times: Union[Iterable[int], None] = None
    ) -> List[Optional[float]]:
        return self._extract_sequential_data_by_time(
            times, self._last_executed_prices, allow_none=True
        )

    def get_last_executed_price(self, time: Union[int, None] = None) -> Optional[float]:
        return self._extract_data_by_time(
            time, self._last_executed_prices, allow_none=True
        )

    def get_fundamental_prices(
        self, times: Union[Iterable[int], None] = None
    ) -> List[float]:
        return cast(
            List[float],
            self._extract_sequential_data_by_time(times, self._fundamental_prices),
        )

    def get_fundamental_price(self, time: Union[int, None] = None) -> float:
        return cast(float, self._extract_data_by_time(time, self._fundamental_prices))

    def get_executed_volumes(
        self, times: Union[Iterable[int], None] = None
    ) -> List[int]:
        return cast(
            List[int],
            self._extract_sequential_data_by_time(
                times, cast(List[Optional[int]], self._executed_volumes)
            ),
        )

    def get_executed_volume(self, time: Union[int, None] = None) -> int:
        return cast(
            int,
            self._extract_data_by_time(
                time, cast(List[Optional[int]], self._executed_volumes)
            ),
        )

    def get_executed_total_prices(
        self, times: Union[Iterable[int], None] = None
    ) -> List[float]:
        return cast(
            List[float],
            self._extract_sequential_data_by_time(
                times, cast(List[Optional[int]], self._executed_total_prices)
            ),
        )

    def get_executed_total_price(self, time: Union[int, None] = None) -> float:
        return cast(
            float,
            self._extract_data_by_time(
                time, cast(List[Optional[int]], self._executed_total_prices)
            ),
        )

    def get_n_buy_orders(self, times: Union[Iterable[int], None] = None) -> List[int]:
        return cast(
            List[int],
            self._extract_sequential_data_by_time(
                times, cast(List[Optional[int]], self._n_buy_orders)
            ),
        )

    def get_n_buy_order(self, time: Union[int, None] = None) -> int:
        return cast(
            int,
            self._extract_data_by_time(
                time, cast(List[Optional[int]], self._n_buy_orders)
            ),
        )

    def get_n_sell_orders(self, times: Union[Iterable[int], None] = None) -> List[int]:
        return cast(
            List[int],
            self._extract_sequential_data_by_time(
                times, cast(List[Optional[int]], self._n_sell_orders)
            ),
        )

    def get_n_sell_order(self, time: Union[int, None] = None) -> int:
        return cast(
            int,
            self._extract_data_by_time(
                time, cast(List[Optional[int]], self._n_sell_orders)
            ),
        )

    def _fill_until(self, time: int) -> None:
        if len(self._mid_prices) >= time + 1:
            return
        length = (time // self.chunk_size + 1) * self.chunk_size
        self._market_prices = self._market_prices + [
            None for _ in range(length - len(self._market_prices))
        ]
        self._mid_prices = self._mid_prices + [
            None for _ in range(length - len(self._mid_prices))
        ]
        self._last_executed_prices = self._last_executed_prices + [
            None for _ in range(length - len(self._last_executed_prices))
        ]
        self._fundamental_prices = self._fundamental_prices + [
            None for _ in range(length - len(self._fundamental_prices))
        ]
        self._executed_volumes = self._executed_volumes + [
            0 for _ in range(length - len(self._executed_volumes))
        ]
        self._executed_total_prices = self._executed_total_prices + [
            0 for _ in range(length - len(self._executed_total_prices))
        ]
        self._n_buy_orders = self._n_buy_orders + [
            0 for _ in range(length - len(self._n_buy_orders))
        ]
        self._n_sell_orders = self._n_sell_orders + [
            0 for _ in range(length - len(self._n_sell_orders))
        ]

    def get_vwap(self, time: Optional[int] = None) -> float:
        if time is None:
            time = self.time
        if time > self.time:
            raise AssertionError("Cannot refer the future parameters")
        if sum(self._executed_volumes[: time + 1]) == 0:
            return float("nan")
        return sum(self._executed_total_prices[: time + 1]) / sum(
            self._executed_volumes[: time + 1]
        )

    @property
    def is_running(self) -> bool:
        return self._is_running

    def get_best_buy_price(self) -> Optional[float]:
        return self.buy_order_book.get_best_price()

    def get_best_sell_price(self) -> Optional[float]:
        return self.sell_order_book.get_best_price()

    def get_sell_order_book(self) -> Dict[Optional[float], int]:
        return self.sell_order_book.get_price_volume()

    def get_buy_order_book(self) -> Dict[Optional[float], int]:
        return self.buy_order_book.get_price_volume()

    def convert_to_tick_level_rounded_lower(self, price: float) -> int:
        return math.floor(price / self.tick_size)

    def convert_to_tick_level_rounded_upper(self, price: float) -> int:
        return math.ceil(price / self.tick_size)

    def convert_to_tick_level(self, price: float, is_buy: bool) -> int:
        if is_buy:
            return self.convert_to_tick_level_rounded_lower(price=price)
        else:
            return self.convert_to_tick_level_rounded_upper(price=price)

    def convert_to_price(self, tick_level: int) -> float:
        return self.tick_size * tick_level

    def _set_time(self, time: int, next_fundamental_price: float) -> None:
        self.time = time
        self.buy_order_book._set_time(time)
        self.sell_order_book._set_time(time)
        self._fill_until(time=time)
        self._fundamental_prices[self.time] = next_fundamental_price
        if self.time > 0:
            executed_prices: List[float] = cast(
                List[float],
                list(
                    filter(
                        lambda x: x is not None, self._last_executed_prices[: self.time]
                    )
                ),
            )
            self._last_executed_prices[self.time] = (
                executed_prices[-1] if sum(executed_prices) > 0 else None
            )
            mid_prices: List[float] = cast(
                List[float],
                list(filter(lambda x: x is not None, self._mid_prices[: self.time])),
            )
            self._mid_prices[self.time] = (
                mid_prices[-1] if sum(mid_prices) > 0 else None
            )
            market_prices: List[float] = cast(
                List[float],
                list(filter(lambda x: x is not None, self._market_prices[: self.time])),
            )
            self._market_prices[self.time] = (
                market_prices[-1] if sum(market_prices) > 0 else None
            )
            if self.is_running:
                if self._last_executed_prices[self.time - 1] is not None:
                    self._market_prices[self.time] = self._last_executed_prices[
                        self.time
                    ]
                elif self._mid_prices[self.time - 1] is not None:
                    self._market_prices[self.time] = self._mid_prices[self.time]

    def _update_time(self, next_fundamental_price: float) -> None:
        self.time += 1
        self.buy_order_book._set_time(self.time)
        self.sell_order_book._set_time(self.time)
        self._fill_until(time=self.time)
        self._fundamental_prices[self.time] = next_fundamental_price
        if self.time > 0:
            self._last_executed_prices[self.time] = self._last_executed_prices[
                self.time - 1
            ]
            self._mid_prices[self.time] = self._mid_prices[self.time - 1]
            self._market_prices[self.time] = self._market_prices[self.time - 1]
            if self.is_running:
                if self._last_executed_prices[self.time - 1] is not None:
                    self._market_prices[self.time] = self._last_executed_prices[
                        self.time - 1
                    ]
                elif self._mid_prices[self.time - 1] is not None:
                    self._market_prices[self.time] = self._mid_prices[self.time - 1]
        else:
            if self._market_prices[self.time] is None:
                self._market_prices[self.time] = next_fundamental_price

    def _cancel_order(self, cancel: Cancel) -> CancelLog:
        if self.market_id != cancel.order.market_id:
            raise ValueError("this cancel order is for a different market")
        if cancel.order.order_id is None or cancel.order.placed_at is None:
            raise ValueError("the order is not submitted before")
        (self.buy_order_book if cancel.order.is_buy else self.sell_order_book).cancel(
            cancel=cancel
        )
        if cancel.placed_at is None:
            raise AssertionError
        self._update_market_price()

        log: Final[CancelLog] = CancelLog(
            order_id=cancel.order.order_id,
            market_id=cancel.order.market_id,
            cancel_time=cancel.placed_at,
            order_time=cancel.order.placed_at,
            agent_id=cancel.order.agent_id,
            is_buy=cancel.order.is_buy,
            kind=cancel.order.kind,
            volume=cancel.order.volume,
            price=cancel.order.price,
            ttl=cancel.order.ttl,
        )
        if self.logger is not None:
            log.read_and_write(logger=self.logger)
        return log

    def _update_market_price(self) -> None:
        best_buy_price: Optional[float] = self.get_best_buy_price()
        best_sell_price: Optional[float] = self.get_best_sell_price()
        if best_buy_price is None or best_sell_price is None:
            pass
        else:
            self._mid_prices[self.time] = (
                (best_sell_price + best_buy_price) / 2.0
                if best_sell_price is not None and best_buy_price is not None
                else None
            )
        if self.is_running:
            if self._last_executed_prices[self.time] is not None:
                self._market_prices[self.time] = self._last_executed_prices[self.time]
            elif self._mid_prices[self.time] is not None:
                self._market_prices[self.time] = self._mid_prices[self.time]

    def _execute_orders(
        self, price: float, volume: int, buy_order: Order, sell_order: Order
    ) -> ExecutionLog:
        if not self.is_running:
            raise AssertionError("market is not running")
        if buy_order.market_id != self.market_id:
            raise ValueError("buy order is not for this market")
        if sell_order.market_id != self.market_id:
            raise ValueError("sell order is not for this market")

        if buy_order.market_id is None:
            raise ValueError("buy order is not submitted yet")
        if sell_order.market_id is None:
            raise ValueError("sell order is not submitted yet")

        if volume <= 0:
            raise AssertionError

        log: Final[ExecutionLog] = ExecutionLog(
            market_id=self.market_id,
            time=self.time,
            buy_agent_id=buy_order.agent_id,
            sell_agent_id=sell_order.agent_id,
            buy_order_id=cast(int, buy_order.order_id),
            sell_order_id=cast(int, sell_order.order_id),
            price=price,
            volume=volume,
        )

        self.buy_order_book.change_order_volume(order=buy_order, delta=-volume)
        self.sell_order_book.change_order_volume(order=sell_order, delta=-volume)

        self._last_executed_prices[self.time] = price
        self._executed_volumes[self.time] += volume
        self._executed_total_prices[self.time] += volume * price
        self._update_market_price()

        # ToDo: Agent modification will be handled in simulator
        if self.logger is not None:
            log.read_and_write(logger=self.logger)
        return log

    def _add_order(self, order: Order) -> OrderLog:
        if order.market_id != self.market_id:
            raise ValueError("order is not for this market")
        (self.buy_order_book if order.is_buy else self.sell_order_book).add(order=order)
        if order.placed_at is None:
            raise AssertionError
        if order.order_id is not None:
            raise ValueError("the order is already submitted")
        order.order_id = self._next_order_id
        self._next_order_id += 1
        self._update_market_price()
        if order.is_buy:
            self._n_buy_orders[self.time] += 1
        else:
            self._n_sell_orders[self.time] += 1

        log: Final[OrderLog] = OrderLog(
            order_id=order.order_id,
            market_id=order.market_id,
            time=order.placed_at,
            agent_id=order.agent_id,
            is_buy=order.is_buy,
            kind=order.kind,
            volume=order.volume,
            price=order.price,
            ttl=order.ttl,
        )
        if self.logger is not None:
            log.read_and_write(logger=self.logger)
        return log

    def remain_executable_orders(self) -> bool:
        if len(self.sell_order_book) == 0:
            return False
        if len(self.buy_order_book) == 0:
            return False
        sell_best: Order = cast(Order, self.sell_order_book.get_best_order())
        buy_best: Order = cast(Order, self.buy_order_book.get_best_order())
        if sell_best.price is not None or buy_best.price is not None:
            if sell_best.price is not None and buy_best.price is not None:
                return sell_best.price <= buy_best.price
            else:
                return True
        else:
            sell_book: Dict[
                Optional[float], int
            ] = self.sell_order_book.get_price_volume()
            buy_book: Dict[
                Optional[float], int
            ] = self.buy_order_book.get_price_volume()
            if None not in sell_book or None not in buy_book:
                raise AssertionError
            if sell_book[None] != buy_book[None]:
                return True
            else:
                sell_book.pop(None)
                buy_book.pop(None)
                return min(list(cast(Dict[float, int], sell_book).keys())) <= max(
                    list(cast(Dict[float, int], buy_book).keys())
                )

    def _execution(self) -> List[ExecutionLog]:
        if not self.remain_executable_orders():
            return []
        pending: List[Tuple[int, Order, Order]] = []
        buy_order_id_tmp: int = 0
        sell_order_id_tmp: int = -1

        buy_order: Order = self.buy_order_book.priority_queue.queue[buy_order_id_tmp]
        sell_order: Order
        buy_order_volume_tmp: int = buy_order.volume
        sell_order_volume_tmp: int = 0
        price: Optional[float] = None
        while True:
            if buy_order_volume_tmp != 0 and sell_order_volume_tmp != 0:
                raise AssertionError
            if buy_order_volume_tmp == 0:
                buy_order_id_tmp += 1
                if buy_order_id_tmp >= len(self.buy_order_book.priority_queue.queue):
                    break
                buy_order = self.buy_order_book.priority_queue.queue[buy_order_id_tmp]
                buy_order_volume_tmp = buy_order.volume
            if sell_order_volume_tmp == 0:
                sell_order_id_tmp += 1
                if sell_order_id_tmp >= len(self.sell_order_book.priority_queue.queue):
                    break
                sell_order = self.sell_order_book.priority_queue.queue[
                    sell_order_id_tmp
                ]
                sell_order_volume_tmp = sell_order.volume
            if (
                buy_order.price is not None
                and sell_order.price is not None
                and buy_order.price < sell_order.price
            ):
                break
            volume = min(buy_order_volume_tmp, sell_order_volume_tmp)
            buy_order_volume_tmp -= volume
            sell_order_volume_tmp -= volume
            if buy_order_volume_tmp < 0:
                raise AssertionError
            if sell_order_volume_tmp < 0:
                raise AssertionError
            if buy_order.price is None or sell_order.price is None:
                if buy_order.price is None and sell_order.price is None:
                    pending.append((volume, buy_order, sell_order))
                else:
                    price = buy_order.price or sell_order.price
                    pending.append((volume, buy_order, sell_order))
            else:
                if buy_order.placed_at == sell_order.placed_at:
                    # ToDo: in the actual market, this doesn't occur
                    price = buy_order.price
                else:
                    price = (
                        buy_order.price
                        if cast(int, buy_order.placed_at)
                        < cast(int, sell_order.placed_at)
                        else sell_order.price
                    )
                pending.append((volume, buy_order, sell_order))
        if price is None:
            raise AssertionError
        logs: List[ExecutionLog] = list(
            map(
                lambda x: self._execute_orders(
                    price=cast(float, price),
                    volume=x[0],
                    buy_order=x[1],
                    sell_order=x[2],
                ),
                pending,
            )
        )
        if self.logger is not None:
            self.logger.bulk_write(logs=cast(List[Log], logs))
        return logs

    def change_fundamental_price(self, scale: float) -> None:
        time: int = self.time
        current_fundamental: float = self.get_fundamental_price(time=time)
        new_fundamental: float = current_fundamental * scale
        self._fundamental_prices[time] = new_fundamental
        self._simulator.fundamentals.prices[self.market_id][time] = new_fundamental
        self._simulator.fundamentals._generated_until = time



import os
import random
import time
from io import TextIOWrapper
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

from ..agents.base import Agent
from ..agents.high_frequency_agent import HighFrequencyAgent
from ..events import EventABC
from ..events import EventHook
from ..index_market import IndexMarket
from ..logs.base import CancelLog
from ..logs.base import ExecutionLog
from ..logs.base import Log
from ..logs.base import Logger
from ..logs.base import MarketStepBeginLog
from ..logs.base import MarketStepEndLog
from ..logs.base import OrderLog
from ..logs.base import SessionBeginLog
from ..logs.base import SessionEndLog
from ..logs.base import SimulationBeginLog
from ..logs.base import SimulationEndLog
from ..market import Market
from ..order import Cancel
from ..order import Order
from ..session import Session
from ..simulator import Simulator
from ..utils.class_finder import find_class
from ..utils.json_extends import json_extends
from .base import Runner


class SequentialRunner(Runner):
    def __init__(
        self,
        settings: Union[Dict, TextIOWrapper, os.PathLike, str],
        prng: Optional[random.Random] = None,
        logger: Optional[Logger] = None,
        simulator_class: Type[Simulator] = Simulator,
    ):
        super().__init__(settings, prng, logger, simulator_class)
        self._pending_setups: List[Tuple[Callable, Dict]] = []

    @staticmethod
    def judge_hft_or_not(agent: Agent) -> bool:
        return isinstance(agent, HighFrequencyAgent)

    def _generate_markets(self, market_type_names: List[str]) -> None:
        i_market = 0
        for name in market_type_names:
            market_settings: Dict = self.settings[name]
            n_markets = 1
            id_from = 0
            id_to = 0
            if "numMarkets" in market_settings:
                n_markets = int(market_settings["numMarkets"])
                id_to = n_markets - 1
            if "from" in market_settings or "to" in market_settings:
                if "from" not in market_settings or "to" not in market_settings:
                    raise ValueError(
                        f"both {name}.from and {name}.to are required in json file if you use"
                    )
                if "numMarkets" in market_settings:
                    raise ValueError(
                        f"{name}.numMarkets and ({name}.from or {name}.to) cannot be used at the same time"
                    )
                n_markets = market_settings["to"] - market_settings["from"]
                id_from = market_settings["from"]
                id_to = market_settings["to"]
            prefix: str
            if "prefix" in market_settings:
                prefix = market_settings["prefix"]
            else:
                prefix = name + ("-" if n_markets > 1 else "")
            market_settings = json_extends(
                whole_json=self.settings,
                parent_name=name,
                target_json=market_settings,
                excludes_fields=["numMarkets", "from", "to", "prefix"],
            )
            if "class" not in market_settings:
                raise ValueError(f"class is not defined for {name}")
            market_class: Type[Market] = find_class(name=market_settings["class"])
            if not issubclass(market_class, Market):
                raise ValueError(
                    f"market class for {name} does not inherit Market class"
                )
            if "fundamentalPrice" in market_settings:
                fundamental_price = float(market_settings["fundamentalPrice"])
            elif "marketPrice" in market_settings:
                fundamental_price = float(market_settings["marketPrice"])
            else:
                raise ValueError(
                    f"fundamentalPrice or marketPrice is required for {name}"
                )
            fundamental_drift: float = 0.0
            if "fundamentalDrift" in market_settings:
                fundamental_price = float(market_settings["fundamentalDrift"])
            fundamental_volatility: float = 0.0
            if "fundamentalVolatility" in market_settings:
                fundamental_volatility = float(market_settings["fundamentalVolatility"])

            for i in range(id_from, id_to + 1):
                market = market_class(
                    market_id=i_market,
                    prng=random.Random(self._prng.randint(0, 2**31)),
                    simulator=self.simulator,
                    logger=self.logger,
                    name=prefix + (str(i) if n_markets != 1 else ""),
                )
                i_market += 1
                self.simulator._add_market(market=market, group_name=name)
                if not isinstance(market, IndexMarket):
                    self.simulator.fundamentals.add_market(
                        market_id=market.market_id,
                        initial=fundamental_price,
                        drift=fundamental_drift,
                        volatility=fundamental_volatility,
                    )
                self._pending_setups.append(
                    (market.setup, {"settings": market_settings})
                )

    def _generate_agents(self, agent_type_names: List[str]) -> None:
        i_agent = 0
        for name in agent_type_names:
            agent_settings: Dict = self.settings[name]
            n_agents = 1
            id_from = 0
            id_to = 0
            if "numAgents" in agent_settings:
                n_agents = int(agent_settings["numAgents"])
                id_to = n_agents - 1
            if "from" in agent_settings or "to" in agent_settings:
                if "from" not in agent_settings or "to" not in agent_settings:
                    raise ValueError(
                        f"both {name}.from and {name}.to are required in json file if you use"
                    )
                if "numAgents" in agent_settings:
                    raise ValueError(
                        f"{name}.numMarkets and ({name}.from or {name}.to) cannot be used at the same time"
                    )
                n_agents = agent_settings["to"] - agent_settings["from"]
                id_from = agent_settings["from"]
                id_to = agent_settings["to"]
            prefix: str
            if "prefix" in agent_settings:
                prefix = agent_settings["prefix"]
            else:
                prefix = name + ("-" if n_agents > 1 else "")
            agent_settings = json_extends(
                whole_json=self.settings,
                parent_name=name,
                target_json=agent_settings,
                excludes_fields=["numAgents", "from", "to", "prefix"],
            )
            if "class" not in agent_settings:
                raise ValueError(f"class is not defined for {name}")
            agent_class: Type[Agent] = find_class(name=agent_settings["class"])
            if not issubclass(agent_class, Agent):
                raise ValueError(
                    f"market class for {name} does not inherit Market class"
                )
            if "markets" not in agent_settings:
                raise ValueError(f"markets is required in {name}")
            accessible_market_names: List[str] = agent_settings["markets"]
            accessible_market_ids: List[int] = sum(
                [
                    list(
                        map(
                            lambda m: m.market_id,
                            self.simulator.markets_group_name2market[x],
                        )
                    )
                    for x in accessible_market_names
                ],
                [],
            )
            for i in range(id_from, id_to + 1):
                agent = agent_class(
                    agent_id=i_agent,
                    prng=random.Random(self._prng.randint(0, 2**31)),
                    simulator=self.simulator,
                    logger=self.logger,
                    name=prefix + (str(i) if n_agents != 1 else ""),
                )
                i_agent += 1
                self.simulator._add_agent(agent=agent, group_name=name)
                self._pending_setups.append(
                    (
                        agent.setup,
                        {
                            "settings": agent_settings,
                            "accessible_markets_ids": accessible_market_ids,
                        },
                    )
                )

    def _set_fundamental_correlation(self) -> None:
        if "fundamentalCorrelations" in self.settings["simulation"]:
            corr_settings: Dict = self.settings["simulation"]["fundamentalCorrelations"]
            for key, value in corr_settings.items():
                if key == "pairwise":
                    if (
                        not isinstance(value, list)
                        or sum([len(x) != 3 for x in value]) > 0
                    ):
                        raise ValueError(
                            "simulation.fundamentalCorrelations.pairwise has invalid format data"
                        )
                    for (market1_name, market2_name, corr) in value:
                        market1 = self.simulator.name2market[market1_name]
                        market2 = self.simulator.name2market[market2_name]
                        self.simulator.fundamentals.set_correlation(
                            market_id1=market1.market_id,
                            market_id2=market2.market_id,
                            corr=float(corr),
                        )
                else:
                    raise NotImplementedError(
                        f"{key} for simulation.fundamentalCorrelations is not supported"
                    )

    def _generate_sessions(self) -> None:
        session_settings: Dict = self.settings["simulation"]["sessions"]
        if not isinstance(session_settings, list):
            raise ValueError("simulation.sessions must be list[dict]")
        i_session = 0
        i_event = 0
        session_start_time: int = 0
        for session_setting in session_settings:
            if "sessionName" not in session_setting:
                raise ValueError(
                    "for each element in simulation.sessions must have sessionName"
                )
            session = Session(
                session_id=i_session,
                prng=random.Random(self._prng.randint(0, 2**31)),
                session_start_time=session_start_time,
                simulator=self.simulator,
                name=session_setting["sessionName"],
                logger=self.logger,
            )
            i_session += 1
            if "iterationSteps" not in session_setting:
                raise ValueError(
                    "iterationSteps is required in each element of simulation.sessions"
                )
            session_start_time += session_setting["iterationSteps"]
            self.simulator._add_session(session=session)
            self._pending_setups.append((session.setup, {"settings": session_setting}))
            if "events" in session_setting:
                event_names: List[str] = session_setting["events"]
                for event_name in event_names:
                    event_setting = self.settings[event_name]
                    event_setting = json_extends(
                        whole_json=self.settings,
                        parent_name=event_name,
                        target_json=event_setting,
                        excludes_fields=["numMarkets", "from", "to", "prefix"],
                    )
                    if "class" not in event_setting:
                        raise ValueError(f"class is required in {event_name}")
                    event_class_name = event_setting["class"]
                    event_class: Type[EventABC] = find_class(name=event_class_name)
                    event = event_class(
                        event_id=i_event,
                        prng=random.Random(self._prng.randint(0, 2**31)),
                        session=session,
                        simulator=self.simulator,
                        name=event_name,
                    )
                    i_event += 1
                    self._pending_setups.append(
                        (event.setup, {"settings": event_setting})
                    )

                    def event_hook_setup(_event: EventABC):
                        event_hooks: List[EventHook] = _event.hook_registration()
                        for event_hook in event_hooks:
                            self.simulator._add_event(event_hook)

                    self._pending_setups.append((event_hook_setup, {"_event": event}))

    def _setup(self) -> None:
        if "simulation" not in self.settings:
            raise ValueError("simulation is required in json file")

        if "markets" not in self.settings["simulation"]:
            raise ValueError("simulation.markets is required in json file")
        market_type_names: List[str] = self.settings["simulation"]["markets"]
        if (
            not isinstance(market_type_names, list)
            or sum([not isinstance(m, str) for m in market_type_names]) > 0
        ):
            raise ValueError("simulation.markets in json file have to be list[str]")
        self._generate_markets(market_type_names=market_type_names)
        self._set_fundamental_correlation()

        if "agents" not in self.settings["simulation"]:
            raise ValueError("agents.markets is required in json file")
        agent_type_names: List[str] = self.settings["simulation"]["agents"]
        if (
            not isinstance(agent_type_names, list)
            or sum([not isinstance(m, str) for m in agent_type_names]) > 0
        ):
            raise ValueError("simulation.agents in json file have to be list[str]")
        self._generate_agents(agent_type_names=agent_type_names)

        if "sessions" not in self.settings["simulation"]:
            raise ValueError("simulation.sessions is required in json file")
        session_settings: List[Dict[str, Any]] = self.settings["simulation"]["sessions"]
        if (
            not isinstance(session_settings, list)
            or sum([not isinstance(m, dict) for m in session_settings]) > 0
        ):
            raise ValueError("simulation.sessions in json file have to be List[Dict]")
        self._generate_sessions()

        _ = [func(**kwargs) for func, kwargs in self._pending_setups]

    def _collect_orders(self, session: Session) -> List[List[Union[Order, Cancel]]]:
        agents = self.simulator.normal_frequency_agents
        agents = self._prng.sample(agents, len(agents))
        n_orders = 0
        all_orders: List[List[Union[Order, Cancel]]] = []
        for agent in agents:
            if n_orders >= session.max_normal_orders:
                break
            orders = agent.submit_orders(markets=self.simulator.markets)
            if len(orders) > 0:
                if not session.with_order_placement:
                    raise AssertionError("currently order is not accepted")
                if sum([order.agent_id != agent.agent_id for order in orders]) > 0:
                    raise ValueError(
                        "spoofing order is not allowed. please check agent_id in order"
                    )
                all_orders.append(orders)
                n_orders += 1
        return all_orders

    def _handle_orders(
        self, session: Session, local_orders: List[List[Union[Order, Cancel]]]
    ) -> List[List[Union[Order, Cancel]]]:
        agents = self.simulator.high_frequency_agents
        agents = self._prng.sample(agents, len(agents))
        sequential_orders = self._prng.sample(local_orders, len(local_orders))
        all_orders: List[List[Union[Order, Cancel]]] = []
        for orders in sequential_orders:
            for order in orders:
                if not session.with_order_placement:
                    raise AssertionError("currently order is not accepted")
                market: Market = self.simulator.id2market[order.market_id]
                if isinstance(order, Order):
                    self.simulator._trigger_event_before_order(order=order)
                    log: OrderLog = market._add_order(order=order)
                    agent: Agent = self.simulator.id2agent[order.agent_id]
                    agent.submitted_order(log=log)
                    self.simulator._trigger_event_after_order(order_log=log)
                elif isinstance(order, Cancel):
                    self.simulator._trigger_event_before_cancel(cancel=order)
                    log_: CancelLog = market._cancel_order(cancel=order)
                    agent = self.simulator.id2agent[order.order.agent_id]
                    agent.canceled_order(log=log_)
                    self.simulator._trigger_event_after_cancel(cancel_log=order)
                else:
                    raise NotImplementedError
                if session.with_order_execution:
                    logs: List[ExecutionLog] = market._execution()
                    for execution_log in logs:
                        agent.executed_order(log=execution_log)
                        self.simulator._trigger_event_after_execution(
                            execution_log=execution_log
                        )

            if session.high_frequency_submission_rate < self._prng.random():
                continue

            n_high_freq_orders = 0
            for agent in agents:
                if n_high_freq_orders >= session.max_high_frequency_orders:
                    break

                high_freq_orders: List[Union[Order, Cancel]] = agent.submit_orders(
                    markets=self.simulator.markets
                )
                if len(high_freq_orders) > 0:
                    if not session.with_order_placement:
                        raise AssertionError("currently order is not accepted")
                    if (
                        sum(
                            [
                                order.agent_id != agent.agent_id
                                for order in high_freq_orders
                            ]
                        )
                        > 0
                    ):
                        raise ValueError(
                            "spoofing order is not allowed. please check agent_id in order"
                        )
                    all_orders.append(high_freq_orders)
                    n_high_freq_orders += 1
                    for order in high_freq_orders:
                        market = self.simulator.id2market[order.market_id]
                        if isinstance(order, Order):
                            self.simulator._trigger_event_before_order(order=order)
                            log = market._add_order(order=order)
                            agent = self.simulator.id2agent[order.agent_id]
                            agent.submitted_order(log=log)
                            self.simulator._trigger_event_after_order(order_log=log)
                        elif isinstance(order, Cancel):
                            self.simulator._trigger_event_before_cancel(cancel=order)
                            log_ = market._cancel_order(cancel=order)
                            agent = self.simulator.id2agent[order.order.agent_id]
                            agent.canceled_order(log=log_)
                            self.simulator._trigger_event_after_cancel(cancel_log=order)
                        else:
                            raise NotImplementedError
                        if session.with_order_execution:
                            logs = market._execution()
                            for execution_log in logs:
                                agent.executed_order(log=execution_log)
                                self.simulator._trigger_event_after_execution(
                                    execution_log=execution_log
                                )
        return all_orders

    def _update_markets(self, session: Session) -> None:
        local_orders: List[List[Union[Order, Cancel]]] = self._collect_orders(
            session=session
        )
        self._handle_orders(session=session, local_orders=local_orders)

    def _iterate_market_updates(self, session: Session) -> None:
        markets: List[Market] = self.simulator.markets
        for market in markets:
            market._is_running = session.with_order_execution

        for _ in range(session.iteration_steps):
            for market in markets:
                self.simulator._trigger_event_before_step_for_market(market=market)
                if self.logger is not None:
                    log: Log = MarketStepBeginLog(
                        session=session, market=market, simulator=self.simulator
                    )
                    log.read_and_write_with_direct_process(logger=self.logger)
            if session.with_order_placement:
                self._update_markets(session=session)
            for market in markets:
                if self.logger is not None:
                    log = MarketStepEndLog(
                        session=session, market=market, simulator=self.simulator
                    )
                    log.read_and_write_with_direct_process(logger=self.logger)
                self.simulator._trigger_event_after_step_for_market(market=market)
            self.simulator._update_times_on_markets(self.simulator.markets)  # t++

    def _run(self) -> None:
        if self.logger is not None:
            log: Log = SimulationBeginLog(simulator=self.simulator)  # must be blocking
            log.read_and_write(logger=self.logger)
            self.logger._process()
        self.simulator._update_times_on_markets(self.simulator.markets)  # t: -1 -> 0

        for session in self.simulator.sessions:
            self.simulator.current_session = session
            self.simulator._trigger_event_before_session(session=session)
            if self.logger is not None:
                log = SessionBeginLog(
                    session=session, simulator=self.simulator
                )  # must be blocking
                log.read_and_write(logger=self.logger)
                self.logger._process()
            self._iterate_market_updates(session=session)
            self.simulator._trigger_event_after_session(session=session)
            if self.logger is not None:
                log = SessionEndLog(
                    session=session, simulator=self.simulator
                )  # must be blocking
                log.read_and_write(logger=self.logger)
                self.logger._process()
        if self.logger is not None:
            log = SimulationEndLog(simulator=self.simulator)  # must be blocking
            log.read_and_write(logger=self.logger)
            self.logger._process()



import random
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import cast

from .agents.base import Agent
from .agents.high_frequency_agent import HighFrequencyAgent
from .events.base import EventABC
from .events.base import EventHook
from .fundamentals import Fundamentals
from .index_market import IndexMarket
from .logs.base import Logger
from .market import Market
from .session import Session


class Simulator:
    def __init__(
        self,
        prng: random.Random,
        logger: Optional[Logger] = None,
        fundamental_class: Type[Fundamentals] = Fundamentals,
    ) -> None:
        self._prng = prng
        self.logger: Optional[Logger] = logger
        if self.logger is not None:
            self.logger._set_simulator(simulator=self)

        self.n_events: int = 0
        self.events: List[EventABC] = []
        self.id2event: Dict[int, EventABC] = {}
        self.event_hooks: List[EventHook] = []
        self.events_dict: Dict[str, Dict[Optional[int], List[EventHook]]] = {
            "order_before": {},
            "order_after": {},
            "cancel_before": {},
            "cancel_after": {},
            "execution_after": {},
            "session_before": {},
            "session_after": {},
            "market_before": {},
            "market_after": {},
        }
        self.name2event: Dict[str, EventABC] = {}

        self.n_agents: int = 0
        self.agents: List[Agent] = []
        self.high_frequency_agents: List[Agent] = []
        self.normal_frequency_agents: List[Agent] = []
        self.id2agent: Dict[int, Agent] = {}
        self.name2agent: Dict[str, Agent] = {}
        self.agents_group_name2agent: Dict[str, List[Agent]] = {}

        self.n_markets: int = 0
        self.markets: List[Market] = []
        self.id2market: Dict[int, Market] = {}
        self.name2market: Dict[str, Market] = {}
        self.markets_group_name2market: Dict[str, List[Market]] = {}

        self.fundamentals = fundamental_class(
            prng=random.Random(self._prng.randint(0, 2**31))
        )

        self.n_sessions: int = 0
        self.sessions: List[Session] = []
        self.id2session: Dict[int, Session] = {}
        self.name2session: Dict[str, Session] = {}
        self.current_session: Optional[Session] = None

    def _add_event(self, event_hook: EventHook) -> None:
        if event_hook in self.event_hooks:
            raise ValueError("event_hook is already registered")
        if event_hook.event in self.events:
            raise ValueError("event is already registered")
        if event_hook.event.event_id in self.id2event:
            raise ValueError(f"event_id {event_hook.event.event_id} is duplicated")
        if event_hook.event.name in self.name2event:
            raise ValueError(f"event name {event_hook.event.name} is duplicate")
        event = event_hook.event
        self.events.append(event)
        self.n_events += 1
        self.id2event[event.event_id] = event
        self.name2event[event.name] = event
        self.event_hooks.append(event_hook)
        register_name: str = event_hook.hook_type + (
            "_before" if event_hook.is_before else "_after"
        )
        times: List[Optional[int]] = (
            cast(List[Optional[int]], event_hook.time)
            if event_hook.time is not None
            else cast(List[Optional[int]], [None])
        )
        for time_ in times:
            if time_ not in self.events_dict[register_name]:
                self.events_dict[register_name][time_] = []
            self.events_dict[register_name][time_].append(event_hook)

    def _add_market(self, market: Market, group_name: Optional[str] = None) -> None:
        if market in self.markets:
            raise ValueError("market is already registered")
        if market.market_id in self.id2market:
            raise ValueError(f"market_id {market.market_id} is duplicated")
        if market.name in self.name2market:
            raise ValueError(f"market name {market.name} is duplicate")
        self.markets.append(market)
        self.n_markets += 1
        self.id2market[market.market_id] = market
        self.name2market[market.name] = market
        if group_name is not None:
            if group_name not in self.markets_group_name2market:
                self.markets_group_name2market[group_name] = []
            self.markets_group_name2market[group_name].append(market)

    def _add_agent(self, agent: Agent, group_name: Optional[str] = None) -> None:
        if agent in self.agents:
            raise ValueError("agent is already registered")
        if agent.agent_id in self.id2agent:
            raise ValueError(f"agent_id {agent.agent_id} is duplicated")
        if agent.name in self.name2agent:
            raise ValueError(f"agent name {agent.name} is duplicate")
        self.agents.append(agent)
        self.n_agents += 1
        self.id2agent[agent.agent_id] = agent
        self.name2agent[agent.name] = agent
        if isinstance(agent, HighFrequencyAgent):
            self.high_frequency_agents.append(agent)
        else:
            self.normal_frequency_agents.append(agent)
        if group_name is not None:
            if group_name not in self.agents_group_name2agent:
                self.agents_group_name2agent[group_name] = []
            self.agents_group_name2agent[group_name].append(agent)

    def _add_session(self, session: Session) -> None:
        if session in self.sessions:
            raise ValueError("session is already registered")
        if session.session_id in self.id2session:
            raise ValueError(f"session_id {session.session_id} is duplicated")
        if session.name in self.name2session:
            raise ValueError(f"session name {session.name} is duplicate")
        self.sessions.append(session)
        self.n_sessions += 1
        self.id2session[session.session_id] = session
        self.name2session[session.name] = session

    def _update_time_on_market(self, market: Market) -> None:
        """be careful index matket have to be update after component markets, Technically, the fundamental values for components markets can be calculated beforehand, but not allowed to avoid future data leakage"""
        if not isinstance(market, IndexMarket):
            market._update_time(
                next_fundamental_price=self.fundamentals.get_fundamental_price(
                    market_id=market.market_id, time=market.get_time() + 1
                )
            )
        else:
            market._update_time(
                next_fundamental_price=market.compute_fundamental_index(
                    time=market.get_time() + 1
                )
            )

    def _update_times_on_markets(self, markets: List[Market]) -> None:
        for market in filter(lambda x: not isinstance(x, IndexMarket), markets):
            self._update_time_on_market(market=market)
        for market in filter(lambda x: isinstance(x, IndexMarket), markets):
            self._update_time_on_market(market=market)

    def _check_event_class_and_instance(
        self,
        check_object: object,
        class_requirement: Optional[Type] = None,
        instance_requirement: Optional[object] = None,
    ) -> bool:
        if class_requirement is not None:
            if not isinstance(check_object, class_requirement):
                return False
        if instance_requirement is not None:
            if instance_requirement != check_object:
                return False
        return True

    def _trigger_event_before_order(self, order: "Order") -> None:  # type: ignore
        time: int = order.placed_at
        event_hooks = self.events_dict["order_before"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_before_order(simulator=self, order=order)

    def _trigger_event_after_order(self, order_log: "OrderLog") -> None:  # type: ignore
        time: int = order_log.time
        event_hooks = self.events_dict["order_after"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_after_order(simulator=self, order_log=order_log)

    def _trigger_event_before_cancel(self, cancel: "Cancel") -> None:  # type: ignore
        time: int = cancel.placed_at
        event_hooks = self.events_dict["cancel_before"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_before_cancel(simulator=self, cancel=cancel)

    def _trigger_event_after_cancel(self, cancel_log: "CancelLog") -> None:  # type: ignore
        time: int = cancel_log.time
        event_hooks = self.events_dict["cancel_after"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_after_cancel(simulator=self, cancel_log=cancel_log)

    def _trigger_event_after_execution(self, execution_log: "ExecutionLog") -> None:  # type: ignore
        time: int = execution_log.time
        event_hooks = self.events_dict["execution_after"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_after_execution(
                simulator=self, execution_log=execution_log
            )

    def _trigger_event_before_session(self, session: "Session") -> None:  # type: ignore
        time: int = session.session_start_time
        event_hooks = self.events_dict["session_before"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_before_session(simulator=self, session=session)

    def _trigger_event_after_session(self, session: "Session") -> None:  # type: ignore
        time: int = session.session_start_time + session.iteration_steps - 1
        event_hooks = self.events_dict["session_after"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_after_session(simulator=self, session=session)

    def _trigger_event_before_step_for_market(self, market: "Market") -> None:  # type: ignore
        time: int = market.get_time()
        event_hooks = self.events_dict["market_before"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            if self._check_event_class_and_instance(
                check_object=market,
                class_requirement=event_hook.specific_class,
                instance_requirement=event_hook.specific_instance,
            ):
                event_hook.event.hooked_before_step_for_market(
                    simulator=self, market=market
                )

    def _trigger_event_after_step_for_market(self, market: "Market") -> None:  # type: ignore
        time: int = market.get_time()
        event_hooks = self.events_dict["market_after"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            if self._check_event_class_and_instance(
                check_object=market,
                class_requirement=event_hook.specific_class,
                instance_requirement=event_hook.specific_instance,
            ):
                event_hook.event.hooked_after_step_for_market(
                    simulator=self, market=market
                )

    # ToDo get_xxx_by_name



import warnings
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import cast

from .market import Market


class IndexMarket(Market):
    _components: List[Market] = []

    def setup(self, settings: Dict[str, Any], *args, **kwargs) -> None:  # type: ignore
        super(IndexMarket, self).setup(settings, *args, **kwargs)
        if "markets" not in settings:
            raise ValueError("markets is required for index markets as components")
        if "requires" in settings:
            warnings.warn("requires in index market settings is no longer required")
        for market_name in settings["markets"]:
            market: Market = self._simulator.name2market[market_name]
            self._add_market(market=market)

    def _add_market(self, market: Market) -> None:
        if market in self._components:
            raise ValueError("market is already registered as components")
        if market.outstanding_shares is None:
            raise AssertionError(
                "outstandingShares is required in component market setting"
            )
        self._components.append(market)

    def _add_markets(self, markets: List[Market]) -> None:
        for market in markets:
            self._add_market(market=market)

    def compute_fundamental_index(self, time: Optional[int] = None) -> float:
        if time is None:
            time = self.get_time()
        total_value: float = 0
        total_shares: int = 0
        for market in self._components:
            outstanding_shares = cast(int, market.outstanding_shares)
            total_value += market.get_fundamental_price(time=time) * outstanding_shares
            total_shares += cast(int, outstanding_shares)
        return total_value / total_shares

    def compute_market_index(self, time: Optional[int] = None) -> float:
        if time is None:
            time = self.get_time()
        total_value: float = 0
        total_shares: int = 0
        for market in self._components:
            outstanding_shares = cast(int, market.outstanding_shares)
            total_value += market.get_market_price(time=time) * outstanding_shares
            total_shares += outstanding_shares
        return total_value / total_shares

    def get_components(self) -> List[Market]:
        return self._components

    def is_all_markets_running(self) -> bool:
        return sum(map(lambda x: not x.is_running, self._components)) == 0

    def get_fundamental_index(self, time: Optional[int] = None) -> float:
        return self.get_fundamental_price(time=time)

    def get_market_index(self, time: Optional[int] = None) -> float:
        return self.compute_market_index(time=time)

    def get_index(self, time: Optional[int] = None) -> float:
        return self.get_market_index(time=time)



from typing import Type


def find_class(name: str) -> Type:
    # ToDo check more import
    _1 = __import__("pams", globals(), locals())
    _2 = __import__("pams.agents", globals(), locals(), ["*"])
    _3 = __import__("pams.events", globals(), locals(), ["*"])

    object_class_candidates = [
        getattr(m, name)
        for m in [*globals().values(), *locals().values()]
        if hasattr(m, name)
    ]
    if len(object_class_candidates) != 1:
        raise AttributeError(
            f"class for {name} is found {len(object_class_candidates)} times"
        )
    object_class: Type = object_class_candidates[0]
    return object_class



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pams',
 'pams.agents',
 'pams.events',
 'pams.logs',
 'pams.runners',
 'pams.utils']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.23.4,<2.0.0', 'scipy>=1.9.3,<2.0.0', 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'pams',
    'version': '0.0.6',
    'description': 'PAMS: Platform for Artificial Market Simulations',
    'long_description': '# pams\nPAMS: Platform for Artificial Market Simulations\n\n[![python](https://img.shields.io/pypi/pyversions/pams.svg)](https://pypi.org/project/pams)\n[![pypi](https://img.shields.io/pypi/v/pams.svg)](https://pypi.org/project/pams)\n[![CI](https://github.com/masanorihirano/pams/actions/workflows/ci-python.yml/badge.svg)](https://github.com/masanorihirano/pams/actions/workflows/ci-python.yml)\n[![downloads](https://img.shields.io/pypi/dm/pams)](https://pypi.org/project/pams)\n[![code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Codacy Badge](https://app.codacy.com/project/badge/Grade/18ed1eecc4f34a99bb6fd9a7160f78ca)](https://www.codacy.com/gh/masanorihirano/pams/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=masanorihirano/pams&amp;utm_campaign=Badge_Grade)\n[![codecov](https://codecov.io/gh/masanorihirano/pams/branch/main/graph/badge.svg?token=tFccElw7Wd)](https://codecov.io/gh/masanorihirano/pams)\n\n## Documentations & User Guides\n\nDocumentations are available on [readthedoc](https://pams.hirano.dev/)\n\n## Install\nThis package is available on pypi as [`pams`](https://pypi.org/project/pams/)\n```bash\n$ pip install pams\n$ python\n>> import pams\n```\n\n## Citation\nCurrently, no publication related to this repository. Therefore, please cite this repository directly if you want.\n```bibtex\n@misc{hirano2022-pams\n    title={{PAMS: Platform for Artificial Market Simulations}},\n    autors={Masanori HIRANO},\n    year={2022},\n    url = {https://github.com/masanorihirano/pams}\n}\n```\n\n## Issues and Contribution\nAbout issues (bugs):\n-   You can report issues [here](https://github.com/masanorihirano/slack_transfer/issues).\n-   There are no guarantee to support or fix those issues.\n\nContributions:\n-   You can send pull requests (PRs) to this repository.\n-   But, there are no guarantee to merge your PRs.\n\n',
    'author': 'Masanori HIRANO',
    'author_email': 'masa.hirano.1996@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import random
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from ..logs.base import CancelLog
from ..logs.base import ExecutionLog
from ..logs.base import Logger
from ..logs.base import OrderLog
from ..market import Market
from ..order import Cancel
from ..order import Order
from ..utils.json_random import JsonRandom


class Agent(ABC):
    """Agent class (abstract class)

    Once you define the agent class inheriting this agent ABC class, simulator automatically generate agents using the class you define.
    Therefore, you don't need to call __init__ or any other method and simulator automatically call them if it is necessary.

    `submit_orders(self, markets: List[Market])` is required to be implemented.

    Note:
        Please also see :class:`pams.agents.FCNAgent`
    """

    def __init__(
        self,
        agent_id: int,
        prng: random.Random,
        simulator: "Simulator",  # type: ignore
        name: str,
        logger: Optional[Logger] = None,
    ) -> None:
        """agent initialization. Usually call from simulator automatically.

        Args:
            agent_id (int): agent id. This is also required to generate orders to identify who submit orders.
            prng (random.Random): pseudo random number generator for this agent.
            simulator (Simulator): this is used for accessing simulation environment.
            name (str): this is unique name for this agent.
                        Automatically generated by simulator and just set it as the name of this agent.
            logger (Logger, Optional): logger for correcting various outputs in one simulation.
                                       logger is usually shared to all classes.
                                       Please note that logger is usually not thread-safe and non-blocking.

        Returns:
            None

        Note:
             `prng` should not be shared with other classed and be used only in this class.
             It is because sometimes agent process runs one of parallelized threads.

        """
        self.agent_id: int = agent_id
        self.name: str = name
        self.asset_volumes: Dict[int, int] = {}
        self.cash_amount: float = 0
        self.prng: random.Random = prng
        self.sim: "Simulator" = simulator  # type: ignore
        self.logger: Optional[Logger] = logger

    def setup(self, settings: Dict[str, Any], accessible_markets_ids: List[int], *args, **kwargs) -> None:  # type: ignore
        if "cashAmount" not in settings:
            raise ValueError("cashAmount is required property of agent settings")
        self.cash_amount = settings["cashAmount"]
        if "assetVolume" not in settings:
            raise ValueError("cashAmount is required property of agent settings")

        for market_id in accessible_markets_ids:
            self.set_market_accessible(market_id=market_id)
            volume = int(
                JsonRandom(prng=self.prng).random(json_value=settings["assetVolume"])
            )
            self.set_asset_volume(market_id=market_id, volume=volume)

    def get_asset_volume(self, market_id: int) -> int:
        if not self.is_market_accessible(market_id=market_id):
            raise AssertionError(f"market {market_id} is not accessible")
        return self.asset_volumes[market_id]

    def get_cash_amount(self) -> float:
        return self.cash_amount

    def get_prng(self) -> random.Random:
        return self.prng

    def is_market_accessible(self, market_id: int) -> bool:
        return market_id in self.asset_volumes

    def submitted_order(self, log: OrderLog) -> None:
        pass

    def executed_order(self, log: ExecutionLog) -> None:
        pass

    def canceled_order(self, log: CancelLog) -> None:
        pass

    def set_asset_volume(self, market_id: int, volume: int) -> None:
        if not self.is_market_accessible(market_id=market_id):
            raise AssertionError(f"market {market_id} is not accessible")
        self.asset_volumes[market_id] = volume

    def set_cash_amount(self, cash_amount: float) -> None:
        self.cash_amount = cash_amount

    def set_market_accessible(self, market_id: int) -> None:
        self.asset_volumes[market_id] = 0

    @abstractmethod
    def submit_orders(self, markets: List[Market]) -> List[Union[Order, Cancel]]:
        pass

    def update_asset_volume(self, market_id: int, delta: int) -> None:
        if not self.is_market_accessible(market_id=market_id):
            raise AssertionError(f"market {market_id} is not accessible")
        self.asset_volumes[market_id] += delta

    def update_cash_amount(self, delta: float) -> None:
        self.cash_amount += delta

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + str(id)
            + ","
            + str(self.cash_amount)
            + ","
            + str(self.asset_volumes)
        )
